"""Path-checking functions."""

from pathlib import Path
from typing import Final

_FILE_VALID_STUDENTS_GIT: Final[str] = "validated_git_urls.yml"
_FILE_ASSESSED_STUDENTS: Final[str] = "assessed_students.yml"
_FOLDER_STUDENT_REPOS: Final[str] = "student_repos"


def check_required_fp_exists(fp_to_check: Path) -> None:
    """Check if file path exists, and if not, raise exception."""
    if not fp_to_check.exists():
        raise FileNotFoundError(f"File {fp_to_check!r} does not exist!")


def get_student_repo_fp(assess_fp: Path, student_id: str, repo_name: str) -> Path:
    """Determine the path to a student repository."""
    return assess_fp.joinpath(_FOLDER_STUDENT_REPOS, student_id, repo_name)


def get_student_repos_fp(assess_fp: Path) -> Path:
    """Determine the path containing all student repositories."""
    return assess_fp.joinpath(_FOLDER_STUDENT_REPOS)


def get_valid_students_git_fp(assess_fp: Path) -> Path:
    """Determine path for valid student Git URLs yaml file."""
    return assess_fp.joinpath(_FILE_VALID_STUDENTS_GIT)


def get_assessed_students_fp(assess_fp: Path) -> Path:
    """Determine path for student assessments yaml file."""
    return assess_fp.joinpath(_FILE_ASSESSED_STUDENTS)



"""The egrader command-line interface."""

import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from typing import Final

from sh import ErrorReturnCode

from .assess import assess
from .cli_lib import (
    OPT_E_LONG,
    OPT_E_OVWR,
    OPT_E_SHORT,
    OPT_E_STOP,
    OPT_E_UPDT,
    CLIArgError,
)
from .fetch import fetch
from .plugin import PluginLoadError, list_plugins
from .plugins.report import report_basic
from .report import report

_ASSESS_FOLDER_ATTR: Final[str] = "assess_folder"
_RULES_FILE_ATTR: Final[str] = "rules_file"
_FOLDER_ASSESS_DEFAULT_PREFIX: Final[str] = "out_"


def main():
    """Function invoked with the `egrader` command."""
    # Create an argument parser
    parser = ArgumentParser(
        description="Exercise Grader",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    # Debug flag for showing full stack traces when errors occur
    parser.add_argument(
        "--debug",
        action="store_true",
        help="show complete exception traceback when an error occurs",
    )

    # Allow for subcommands
    subparsers = parser.add_subparsers(title="commands", required=True)

    # Create the parser for the "fetch" command
    parser_fetch = subparsers.add_parser("fetch", help="fetch all repositories")
    parser_fetch.add_argument(
        f"-{OPT_E_SHORT}",
        f"--{OPT_E_LONG}",
        choices=[OPT_E_STOP, OPT_E_UPDT, OPT_E_OVWR],
        help="action to take when fetch has already been performed before "
        f"(default: {OPT_E_STOP})",
        default=OPT_E_STOP,
    )

    parser_fetch.add_argument(
        "urls_file",
        metavar="URLS",
        help="student public Git account URLs file in TSV format",
    )
    parser_fetch.add_argument(
        _RULES_FILE_ATTR,
        metavar=_RULES_FILE_ATTR.upper(),
        help="assessment rules in YAML format",
    )
    parser_fetch.add_argument(
        _ASSESS_FOLDER_ATTR,
        metavar=_ASSESS_FOLDER_ATTR.upper(),
        help="Folder where assessment data will be placed (defaults to RULES "
        "minus yaml extension)",
        nargs="?",
    )
    parser_fetch.set_defaults(func=fetch)

    # Create the parser for the "assess" command
    parser_assess = subparsers.add_parser("assess", help="perform assessment")
    parser_assess.add_argument(
        _RULES_FILE_ATTR,
        metavar=_RULES_FILE_ATTR.upper(),
        help="assessment rules in YAML format",
    )
    parser_assess.add_argument(
        _ASSESS_FOLDER_ATTR,
        metavar=_ASSESS_FOLDER_ATTR.upper(),
        help="Folder where assessment data is located (defaults to RULES "
        "minus yaml extension)",
        nargs="?",
    )
    parser_assess.set_defaults(func=assess)

    # Create the parser for the "report" command
    parser_report = subparsers.add_parser(
        "report", help="generate an assessment report"
    )
    parser_report.add_argument(
        _ASSESS_FOLDER_ATTR,
        metavar=_ASSESS_FOLDER_ATTR.upper(),
        help="Folder where assessment data is located",
    )
    default_report = report_basic.__name__[len(report.__name__) + 1 :]
    parser_report.add_argument(
        "report_type",
        metavar="REPORT_TYPE [REPORT_ARGS ...]",
        help=f"Type of report and report arguments (defaults to {default_report})",
        default=default_report,
        nargs="?",
    )
    parser_report.set_defaults(func=report)

    # Create the parser for the "plugins" command
    parser_plugins = subparsers.add_parser("plugins", help="list available plugins")
    parser_plugins.set_defaults(func=list_plugins)

    # Parse command line arguments
    args = parser.parse_known_args()

    # Determine assessment folder
    assess_folder = getattr(args[0], _ASSESS_FOLDER_ATTR, None)
    rules_file = getattr(args[0], _RULES_FILE_ATTR, None)

    if assess_folder is None and rules_file is None and args[0].func != list_plugins:
        # This should not be possible
        raise AssertionError(
            f"{_ASSESS_FOLDER_ATTR.upper()} and {_RULES_FILE_ATTR.upper()} "
            "parameters cannot be None simultaneously!"
        )

    # Assessment folder path can be given by user or obtained from the rules file name
    if assess_folder is not None:
        assess_fp = Path(assess_folder)
    elif rules_file is not None:
        assess_fp = Path(f"{_FOLDER_ASSESS_DEFAULT_PREFIX}{Path(rules_file).stem}")
    else:
        assess_fp = None

    # Invoke function to perform selected command
    try:
        with StringIO() as out_stream, redirect_stdout(out_stream):
            args[0].func(assess_fp, args[0], args[1])
            out_string = out_stream.getvalue()
    except (
        ErrorReturnCode,
        FileNotFoundError,
        FileExistsError,
        CLIArgError,
        PluginLoadError,
        SyntaxError,
    ) as e:
        print(e.args[0], file=sys.stderr)
        if args[0].debug:
            print("-------- Exception details --------", file=sys.stderr)
            raise e
        return 1
    else:
        print(out_string, end="")
        return 0



"""Assessment functions."""

from argparse import Namespace
from pathlib import Path
from typing import Any, Dict, List, MutableSet, Sequence

from .cli_lib import check_empty_args
from .paths import (
    check_required_fp_exists,
    get_assessed_students_fp,
    get_student_repos_fp,
    get_valid_students_git_fp,
)
from .plugin import (
    get_short_plugin_desc,
    load_inter_repo_plugin_functions,
    load_repo_plugin_functions,
)
from .types import AssessedRepo, AssessedStudent, Assessment
from .yaml import load_yaml, save_yaml


def assess(assess_fp: Path, args: Namespace, extra_args: Sequence[str]) -> None:
    """Perform student assessment."""
    # extra_args should be empty
    check_empty_args(extra_args)

    # Check if assessment folder exists, and if not, quit
    check_required_fp_exists(assess_fp)

    # Determine rules file path
    rules_fp: Path = Path(args.rules_file)

    # Check if rules file exists, and if not, quit
    check_required_fp_exists(rules_fp)

    # Determine file path for valid students Git URL and repos yaml file
    students_git_fp: Path = get_valid_students_git_fp(assess_fp)

    # Check if valid students Git URL yaml file exists, and if not, quit
    check_required_fp_exists(students_git_fp)

    # Load rules
    rules = load_yaml(rules_fp)

    # Load student list and their URLs
    students_git = load_yaml(students_git_fp, safe=False)

    # Obtained all the repository assessments defined by the rules
    required_assessments: MutableSet[str] = {
        assess_rule["name"]
        for rule in rules
        if "assessments" in rule
        for assess_rule in rule["assessments"]
    }

    # Load required assessment plugins as specified by the rules
    assess_functions: Dict[str, Any] = load_repo_plugin_functions(required_assessments)

    # Initialize student grades list
    assessed_students: List[AssessedStudent] = []

    # Initialize dictionary of assessed repositories by name, which will be
    # required later for inter-repository assessments
    repos_by_name: Dict[str, List[AssessedRepo]] = {rule["repo"]: [] for rule in rules}

    # Apply rules and assessments to each student
    for student_git in students_git:
        # Create instance of current student's assessment
        assessed_student: AssessedStudent = AssessedStudent(student_git.sid)

        # Loop through rules
        for rule in rules:
            # Create an instance of the repository being assessed
            assessed_repo: AssessedRepo = AssessedRepo(rule["repo"], rule["weight"])

            # If student has the repository specified in the current rule, apply
            # the specified assessments
            if rule["repo"] in student_git.repos:
                # Get the student's repository local path
                assessed_repo.local_path = student_git.repos[rule["repo"]]

                # Append repository being assessed to dictionary of repositories
                # by name (will be required later for inter-repository assessments)
                repos_by_name[rule["repo"]].append(assessed_repo)

                # Loop through the assessments to be made for the current rule's
                # repository, if any
                if "assessments" in rule:
                    for assess_rule in rule["assessments"]:
                        # Get the plugin function which will perform the assessment
                        # and the respective parameters
                        assess_fun = assess_functions[assess_rule["name"]]
                        assess_params = assess_rule.get("params", {})

                        # Perform assessment and obtain the assessment's grade
                        # between 0 and 1
                        assess_grade = assess_fun(
                            assessed_repo.local_path, **assess_params
                        )

                        # Create assessment object
                        assessment = Assessment(
                            assess_rule["name"],
                            get_short_plugin_desc(assess_fun),
                            assess_params,
                            assess_rule["weight"],
                            assess_grade,
                        )

                        # Add it to the repository currently being assessed
                        assessed_repo.add_assessment(assessment)

            # Add assessed repo to student being assessed
            assessed_student.add_assessed_repo(assessed_repo)

        # Add assessed student to list of assessed students
        assessed_students.append(assessed_student)

    # Obtained all the repository assessments defined by the rules
    required_inter_assessments: MutableSet[str] = {
        inter_assess_rule["name"]
        for rule in rules
        if "inter_assessments" in rule
        for inter_assess_rule in rule["inter_assessments"]
    }

    # Load required inter-assessment plugins as specified by the rules
    inter_assess_functions: Dict[str, Any] = load_inter_repo_plugin_functions(
        required_inter_assessments
    )

    # Apply intra-repository assessments
    for rule in rules:
        if "inter_assessments" in rule:
            for inter_assess_rule in rule["inter_assessments"]:
                repos_with_name: List[AssessedRepo] = repos_by_name[rule["repo"]]

                inter_assess_fun = inter_assess_functions[inter_assess_rule["name"]]
                inter_assess_params = inter_assess_rule.get("params", {})

                # Perform inter-repo assessment and obtain the assessment's
                # grade between 0 and 1
                inter_assess_grades = inter_assess_fun(
                    [sr.local_path for sr in repos_with_name], **inter_assess_params
                )

                # Create assessments (one per repos with the current name)
                assessments = [
                    Assessment(
                        inter_assess_rule["name"],
                        get_short_plugin_desc(inter_assess_fun),
                        inter_assess_params,
                        inter_assess_rule["weight"],
                        iag,
                    )
                    for iag in inter_assess_grades
                ]

                # Add assessments to each repo with the current name
                for ar, a in zip(repos_with_name, assessments, strict=True):
                    ar.add_inter_assessment(a)

    # Determine file path where to save assessment information
    assessed_students_fp: Path = get_assessed_students_fp(assess_fp)

    # Save list of assessed students to yaml file
    save_yaml(assessed_students_fp, assessed_students)

    # Number of assessments performed
    n_assessments = sum([s.assessment_count for s in assessed_students])

    # Provide feedback to the user
    print(f"- Absolute assessment path: {assess_fp.absolute()}.")
    print(f"- Fetched student repository information from {students_git_fp}")
    print(
        f"- Performed {n_assessments} assessments on {len(assessed_students)} "
        f"student repositories at {get_student_repos_fp(assess_fp)}."
    )
    print(f"- Updated {assessed_students_fp}.")



"""Classes used in egrader."""

from typing import Any, Dict, List

import requests
import validators


class StudentGit:
    """A student and his Git repositories."""

    def __init__(self, sid: str, email: str, url: str) -> None:
        """Initialize an instance of this class."""
        # Set instance variables
        self.sid: str = sid
        self.email: str = email
        self.url: str = url
        self.valid_url: bool = False
        self.repos: Dict[str, str] = {}

        # Validate URL if it's is well-formed and if it exists (200)
        if validators.url(self.url) and requests.head(self.url).status_code < 400:
            self.valid_url = True

    def __repr__(self) -> str:
        """String representation of this instance for YAML serialization."""
        return "%s(sid=%r, email=%r, url=%r, valid_url=%r, repos=%r)" % (
            self.__class__.__name__,
            self.sid,
            self.email,
            self.url,
            self.valid_url,
            self.repos,
        )

    def add_repo(self, repo_name: str, repo_path: str) -> None:
        """Add a new repository to this student instance."""
        self.repos[repo_name] = repo_path

    @property
    def repo_count(self) -> int:
        """Number of repositories in this student instance."""
        return len(self.repos)


class Assessment:
    """An already performed assessment."""

    def __init__(
        self,
        name: str,
        description: str,
        parameters: Dict[str, Any],
        weight: float,
        grade_raw: float,
    ) -> None:
        """Initialize an instance of this class."""
        # Set instance variables
        self.name: str = name
        self.description: str = description
        self.parameters: Dict[str, Any] = parameters
        self.weight: float = weight
        self.grade_raw: float = grade_raw

    def __repr__(self) -> str:
        """String representation of this instance for YAML serialization."""
        return "%s(name=%r, description=%r, weight=%r, grade_raw=%r)" % (
            self.__class__.__name__,
            self.name,
            self.description,
            self.weight,
            self.grade_raw,
        )

    @property
    def grade_final(self) -> float:
        """Final grade for this assessment."""
        return self.grade_raw * self.weight


class AssessedRepo:
    """An assessed student repository."""

    def __init__(self, name: str, weight: float) -> None:
        """Initialize an instance of this class."""
        # Set instance variables
        self.name: str = name
        self.weight: float = weight
        self.assessments: List[Assessment] = []
        self.inter_assessments: List[Assessment] = []
        self.local_path: str | None = None

    def __repr__(self) -> str:
        """String representation of this instance for YAML serialization."""
        return "%s(name=%r, weight=%r, assessments=%r, inter_assessments=%r)" % (
            self.__class__.__name__,
            self.name,
            self.weight,
            self.assessments,
            self.inter_assessments,
        )

    def add_assessment(self, assessment: Assessment) -> None:
        """Add an assessment to this repository."""
        self.assessments.append(assessment)

    def add_inter_assessment(self, assessment: Assessment) -> None:
        """Add an inter-repository assessment to this repository."""
        self.inter_assessments.append(assessment)

    def is_empty(self) -> bool:
        """Does this repository have any assessments?"""
        return len(self.assessments) + len(self.inter_assessments) == 0

    @property
    def grade_final(self) -> float:
        """Final grade for this repository."""
        return self.grade_raw * self.weight

    @property
    def grade_raw(self) -> float:
        """Raw grade for this repository."""
        return sum([a.grade_final for a in self.assessments]) + sum(
            [a.grade_final for a in self.inter_assessments]
        )

    @property
    def assessment_count(self) -> int:
        """Number of assessment performed in this repository."""
        return len(self.assessments) + len(self.inter_assessments)


class AssessedStudent:
    """An assessed student."""

    def __init__(self, sid: str) -> None:
        """Initialize an instance of this class."""
        # Set instance variables
        self.sid: str = sid
        self.assessed_repos: List[AssessedRepo] = []

    def __repr__(self) -> str:
        """String representation of this instance for YAML serialization."""
        return "%s(sid=%r, assessed_repos=%r)" % (
            self.__class__.__name__,
            self.sid,
            self.assessed_repos,
        )

    def add_assessed_repo(self, assessed_repo: AssessedRepo) -> None:
        """Add an assessed repository to this student."""
        self.assessed_repos.append(assessed_repo)

    @property
    def grade(self) -> float:
        """This student's grade."""
        return sum([r.grade_final for r in self.assessed_repos])

    @property
    def assessment_count(self) -> int:
        """Number of assessments performed for this student."""
        return sum([r.assessment_count for r in self.assessed_repos])



"""YAML handling functionality."""

from pathlib import Path
from typing import Any

import yaml


def load_yaml(yaml_fp: Path, safe: bool = True) -> Any:
    """Load a yaml file."""
    try:
        with open(yaml_fp, "r") as yaml_file:
            if safe:
                yaml_obj = yaml.safe_load(yaml_file)
            else:
                yaml_obj = yaml.load(yaml_file, yaml.CLoader)
    except yaml.scanner.ScannerError as se:
        raise SyntaxError(
            f"Syntax error{se.problem_mark} {se.context}: {se.problem}"
        ) from se

    return yaml_obj


def save_yaml(yaml_fp: Path, data: Any) -> None:
    """Save a yaml file."""
    yaml_text = yaml.dump(data, Dumper=yaml.CDumper)
    with open(yaml_fp, "w") as yaml_file:
        print(yaml_text, file=yaml_file)



"""Reporting functionality."""

from argparse import Namespace
from pathlib import Path
from typing import Sequence, cast

from .paths import check_required_fp_exists, get_assessed_students_fp
from .plugin import load_report_plugin_function
from .types import AssessedStudent
from .yaml import load_yaml


def report(assess_fp: Path, args: Namespace, extra_args: Sequence[str]) -> None:
    """Generate an assessment report."""
    # Determine report file path
    assess_file_fp: Path = get_assessed_students_fp(assess_fp)

    # Check if assessment file exists, and if not, quit
    check_required_fp_exists(assess_file_fp)

    # Load report file
    assessed_students: Sequence[AssessedStudent] = cast(
        Sequence[AssessedStudent], load_yaml(assess_file_fp, False)
    )

    # Load plugin function to perform reporting
    report_fun = load_report_plugin_function(args.report_type)

    # Invoke reporting function with the specified arguments, if any
    out_string = report_fun(assess_fp, assessed_students, extra_args)

    # Show report stdout output, if any
    if out_string is not None:
        print(out_string, end="")



"""Functions for fetching code from student repositories."""

import shutil
from argparse import Namespace
from pathlib import Path
from typing import List, Sequence

from yarl import URL

from .cli_lib import OPT_E_LONG, OPT_E_OVWR, OPT_E_SHORT, OPT_E_STOP, check_empty_args
from .git import GitError, git, git_at
from .paths import (
    check_required_fp_exists,
    get_student_repo_fp,
    get_student_repos_fp,
    get_valid_students_git_fp,
)
from .types import StudentGit
from .yaml import load_yaml, save_yaml


def fetch(assess_fp: Path, args: Namespace, extra_args: Sequence[str]) -> None:
    """Fetch operation: verify Git URLs, clone or update all repositories."""
    print(f"- Absolute assessment path: {assess_fp.absolute()}.")

    # extra_args should be empty
    check_empty_args(extra_args)

    # Determine file paths for Git URLs and rules files
    urls_fp: Path = Path(args.urls_file)
    rules_fp: Path = Path(args.rules_file)

    # Check if Git URLs file exists, and if not, quit
    check_required_fp_exists(urls_fp)

    # Check if rules file exists, and if not, quit
    check_required_fp_exists(rules_fp)

    # Determine file path for validated URLs yaml file
    students_git_fp: Path = get_valid_students_git_fp(assess_fp)

    # Check if the assessment output folder exists
    if assess_fp.exists():
        # If so, action to take depends on the -e command line option
        if getattr(args, OPT_E_LONG) == OPT_E_STOP:
            # Stop processing
            raise FileExistsError(
                "Assessment folder already exists, stopping operation. Check the "
                f"-{OPT_E_SHORT}/--{OPT_E_LONG} option for alternative behavior."
            )

        elif getattr(args, OPT_E_LONG) == OPT_E_OVWR:
            # Delete folder and its contents and recreate it
            print(f"- Assessment folder already exists at {assess_fp}, deleting it.")
            shutil.rmtree(assess_fp)
            assess_fp.mkdir()
    else:
        # Folder doesn't exist, create it
        assess_fp.mkdir()

    # Load rules
    repo_rules = load_yaml(rules_fp)

    # Declare list of student valid Git URLs
    students_git: List[StudentGit]

    # Load student Git URLs
    if students_git_fp.exists():
        # If file with validated URLs already exists, load info from there to
        # avoid rechecking the URLs (only with "-e update" option)
        students_git = load_yaml(students_git_fp, safe=False)

    else:
        # Otherwise load info from original file and validate URLs
        students_git = load_urls(urls_fp)

    # Clone or update student repositories
    n_valid_urls = fetch_repos(
        assess_fp, students_git, [rule["repo"] for rule in repo_rules]
    )

    # Determine number of repositories
    n_repos = sum([s.repo_count for s in students_git])

    # Save validated URLs and repositories to avoid rechecking them later with
    # the "-e update" option
    save_yaml(students_git_fp, students_git)

    # Provide feedback to the user
    print(
        f"- Fetched {n_repos} repositories from {len(students_git)} students, "
        f"{n_valid_urls} of which with valid URLs."
    )
    print(f"- Repositories saved at {get_student_repos_fp(assess_fp)}.")
    print(f"- URL and repository validation report available at {students_git_fp}.")


def fetch_repos(
    assess_fp: Path, students_git: Sequence[StudentGit], repos: Sequence[str]
) -> int:
    """Clone or update student repositories."""
    # Number of valid Git URLs
    n_valid_urls = 0

    # Loop through students
    for student_git in students_git:
        if student_git.valid_url:
            n_valid_urls += 1

            # Loop through mandated repos
            for repo_name in repos:
                # Determine repo URL and local path
                repo_url: URL = URL(student_git.url) / repo_name
                repo_fp: Path = get_student_repo_fp(
                    assess_fp, student_git.sid, repo_name
                )

                # Does the repository already exist?
                if repo_fp.exists():
                    # Path exists, only update repository
                    git_at(repo_fp, "pull")

                    # Add repo location to student object
                    student_git.add_repo(repo_name, str(repo_fp))

                else:
                    # Repository doesn't exist, do a full clone
                    try:
                        git("clone", repo_url, repo_fp)

                    except GitError:
                        # If a GitException occurs, assume the repo doesn't exist
                        pass

                    else:
                        # Otherwise add repo location to student object
                        student_git.add_repo(repo_name, str(repo_fp))

    return n_valid_urls


def load_urls(urls_fp: Path) -> List[StudentGit]:
    """Load student Git URLs."""
    # The student list, initially empty
    students_git: List[StudentGit] = []

    # Open Git URLs file
    with open(urls_fp) as urls_file:
        # Cycle through each line of the file
        for lno, line in enumerate(urls_file, 1):
            # Remove leading and trailing whitespace
            line = line.strip()

            # Ignore line if it's empty or starts with # (comment)
            if len(line) == 0 or line[0] == "#":
                continue

            # Split line and check that it's composed of three chunks
            std_url = line.split()
            if len(std_url) != 3:
                raise SyntaxError(
                    f"Syntax error in line {lno} of {urls_fp}: " f"{line!r}"
                )

            # Create a student instance with it's ID and URL, taken from the line
            student_git = StudentGit(std_url[0], std_url[1], std_url[2])

            # Append new student to the student list
            students_git.append(student_git)

    # Return students
    return students_git



"""Functions for handling Git functionality."""

from sh import ErrorReturnCode
from sh import git as sh_git


class GitError(Exception):
    """Error raised when a Git command fails."""


def git(*args):
    """Run git with the specified arguments."""
    try:
        return sh_git("--no-pager", *args)
    except ErrorReturnCode as erc:
        raise GitError(
            f"The following error occurred when executing the {erc.full_cmd!r} command:"
            f"\n\n{erc.stderr.decode('UTF-8')}"
        ) from erc


def git_at(repo_path, *args):
    """Run git at location given by repo_path with the specified arguments."""
    return git("-C", repo_path, *args)



"""Plug-in handling functionality."""

from argparse import Namespace
from importlib.metadata import EntryPoints, entry_points
from inspect import getdoc
from pathlib import Path
from typing import AbstractSet, Any, Dict, Final, Sequence

from .cli_lib import check_empty_args

_PLUGINS_ASSESS_REPO: Final[str] = "egrader.assess_repo"
_PLUGINS_ASSESS_INTER_REPO: Final[str] = "egrader.assess_inter_repo"
_PLUGINS_REPORT: Final[str] = "egrader.report"


def _load_plugin_functions(
    plugin_group: str, required: AbstractSet[str]
) -> Dict[str, Any]:
    """Load required plugins from the specified plugin group."""
    # Load plugins
    plugins: EntryPoints = entry_points(group=plugin_group)

    # Set of existing plugin names
    plugin_names: AbstractSet[str] = {plugin.name for plugin in plugins}

    # Are there any required plugins not in the existing plugins set?
    plugins_not_found: AbstractSet[str] = required - plugin_names
    if len(plugins_not_found) > 0:
        raise PluginLoadError(f"Required plugins {plugins_not_found} not found.")

    # Load required plugins
    plugin_functions: Dict[str, Any] = {}
    for plugin in plugins:
        if plugin.name in required:
            plugin_functions[plugin.name] = plugin.load()

    return plugin_functions


def _load_plugin_function(plugin_group: str, plugin_name: str) -> Any:
    """Load a specific plugin function."""
    # Obtain the specified plugin entry point
    plugins: EntryPoints = entry_points(group=plugin_group, name=plugin_name)

    # If plugin no found, raise error
    if len(plugins) == 0:
        raise PluginLoadError(
            f"Plugin {plugin_name!r} not found in {plugin_group!r} group"
        )

    # Otherwise, return plugin function
    return plugins[0].load()


class PluginLoadError(Exception):
    """Error raised when a required plugin fails to load."""


def get_short_plugin_desc(func) -> str:
    """Get a short description of a plugin."""
    desc: str | None = getdoc(func)

    if desc is not None and len(desc) > 0:
        desc = desc.split("\n")[0]
    else:
        desc = "Unavailable"

    return desc


def load_repo_plugin_functions(required: AbstractSet[str]) -> Dict[str, Any]:
    """Load required plugins from the intra-repository plugin group."""
    return _load_plugin_functions(_PLUGINS_ASSESS_REPO, required)


def load_inter_repo_plugin_functions(required: AbstractSet[str]) -> Dict[str, Any]:
    """Load required plugins from the inter-repository plugin group."""
    return _load_plugin_functions(_PLUGINS_ASSESS_INTER_REPO, required)


def load_report_plugin_function(plugin_name: str) -> Any:
    """Load a report plugin function."""
    return _load_plugin_function(_PLUGINS_REPORT, plugin_name)


def list_plugins(assess_fp: Path, args: Namespace, extra_args: Sequence[str]):
    """List available plugins."""
    # extra_args should be empty
    check_empty_args(extra_args)

    plugin_types = (
        (":: Repository assessment plugins", _PLUGINS_ASSESS_REPO),
        (":: Inter-repository assessment plugins", _PLUGINS_ASSESS_INTER_REPO),
        (":: Reporting plugins", _PLUGINS_REPORT),
    )

    for plugin_type in plugin_types:
        plugins: EntryPoints = entry_points(group=plugin_type[1])
        print(f"{plugin_type[0]}\n")
        for plugin in plugins:
            print(f"\t{plugin.name}\n\t\t{get_short_plugin_desc(plugin.load())}")
        print()



"""Functions used by the command-line interface."""

from typing import Final, Sequence

OPT_E_SHORT: Final[str] = "e"
OPT_E_LONG: Final[str] = "existing"
OPT_E_STOP: Final[str] = "stop"
OPT_E_UPDT: Final[str] = "update"
OPT_E_OVWR: Final[str] = "overwrite"


class CLIArgError(Exception):
    """Error raised when command-line arguments are invalid."""


def check_empty_args(args: Sequence[str]) -> None:
    """Check that argument list is empty, otherwise raise error."""
    if len(args) > 0:
        raise CLIArgError(None, f"Invalid arguments: {', '.join(args)}")
